module Day2Lib
  ( Opponent (..),
    Me (..),
    Trick (..),
    Hint (..),
    assembleNaiveTrick,
    evaluateTrick,
    evaluateWin,
    evaluateMe,
    parseHint,
    assembleTrick,
    --
    processInput1,
    processInput2,
  )
where

data Opponent = A | B | C deriving (Show, Read, Eq, Enum)

data Me = X | Y | Z deriving (Show, Read, Eq, Enum)

data Hand = R | P | S deriving (Show, Eq, Enum)

newtype Hint = Hint (Opponent, Me) deriving (Show, Eq)

newtype Trick = Trick (Hand, Hand) deriving (Show, Eq)

parseHint :: String -> (Opponent, Me)
parseHint (o : _ : m : []) = (read [o], read [m])

opponentHand :: Opponent -> Hand
opponentHand = toEnum . fromEnum

naiveResponse :: Hand -> Me -> Hand
naiveResponse _ m = toEnum $ fromEnum m

findResponse :: Hand -> Me -> Hand
findResponse o m = toEnum ((fromEnum o + fromEnum m - 1) `mod` 3)

assembleTrick :: Hint -> Trick
assembleTrick (Hint (o, m)) = Trick (opponentHand o, findResponse (opponentHand o) m)

assembleNaiveTrick :: Hint -> Trick
assembleNaiveTrick (Hint (o, m)) = Trick (opponentHand o, naiveResponse (opponentHand o) m)

evaluateTrick :: Trick -> Int
evaluateTrick t = evaluateWin t + evaluateMe t

evaluateMe :: Trick -> Int
evaluateMe (Trick (_, m)) = fromEnum m + 1

evaluateWin :: Trick -> Int
evaluateWin (Trick (o, m)) = 6 - (fromEnum o - fromEnum m + 1) `mod` 3 * 3

-- ######################################

processInput1 :: String -> Int
processInput1 = sum . (map $ evaluateTrick . assembleNaiveTrick . Hint . parseHint) . lines

processInput2 :: String -> Int
processInput2 = sum . (map $ evaluateTrick . assembleTrick . Hint . parseHint) . lines
