module Day9Lib
  ( Rope (..),
    newRope,
    follow,
    --
    processInput1,
    processInput2,
  )
where

import Data.List
import Debug.Trace

newtype Vector = Vector (Int, Int) deriving (Show, Eq)

instance Num Vector where
  Vector v1 + Vector v2 = Vector (fst v1 + fst v2, snd v1 + snd v2)

  Vector v1 - Vector v2 = Vector (fst v1 - fst v2, snd v1 - snd v2)

  Vector v1 * Vector v2 = Vector (fst v1 * fst v2, snd v1 * snd v2)

  fromInteger i = Vector (fromInteger i, fromInteger i)

  abs (Vector v) = Vector (abs . fst $ v, abs . snd $ v)

  signum (Vector (x, y)) = Vector (signum x, signum y)

newtype Rope = Rope [Vector] deriving (Show, Eq)

infix 7 .*

(.*) :: Int -> Vector -> Vector
m .* v = (fromInteger . toInteger) m * v

vlinf :: Vector -> Int
vlinf v = let Vector (ax, ay) = abs v in max ax ay

vtruncate :: Vector -> Vector
vtruncate v = let Vector (ax, ay) = abs v in signum v * Vector (min 1 ax, min 1 ay)

newRope :: Int -> Rope
newRope l = Rope $ replicate l (Vector (0, 0))

move :: Vector -> Char -> Int -> Vector
move v 'L' d = v + d .* Vector (-1, 0)
move v 'R' d = v + d .* Vector (1, 0)
move v 'U' d = v + d .* Vector (0, 1)
move v 'D' d = v + d .* Vector (0, -1)
move v dir d = error $ "wrong direction:" ++ show dir

follow :: Vector -> Vector -> [Vector] -> (Vector, [Vector])
follow h t hist
  | vlinf (h - t) >= 2 =
      let newt = (t + vtruncate (h - t))
       in follow h newt (newt : hist)
  | otherwise = (t, hist)

-- ####################

moveRope :: Rope -> [Vector] -> (Rope, [Vector])
moveRope (Rope [h, t]) hist =
  let (newt, newhist) = follow h t hist
   in (Rope [h, newt], newhist)
moveRope (Rope (h : k1 : ks)) hist =
  let (newk1, _) = follow h k1 []
      (Rope (newks), newhist) = moveRope (Rope (newk1 : ks)) hist
   in (Rope (h : newks), newhist)
moveRope _ _ = error "rope too short"

-- ####################

processInput1 :: String -> Int
processInput1 = length . nub . snd . foldl processLine1 (newRope 2, [Vector (0, 0)]) . lines

processLine1 :: (Rope, [Vector]) -> String -> (Rope, [Vector])
processLine1 (r@(Rope (h : t : [])), hist) (dir : _ : dist) =
  last . take (read dist + 1) . iterate (doStep dir) $ (r, hist)
processLine1 _ s = error $ "processline1 error: " ++ show s

processInput2 :: String -> Int
processInput2 = length . nub . snd . foldl processLine2 (newRope 10, [Vector (0, 0)]) . lines

processLine2 :: (Rope, [Vector]) -> String -> (Rope, [Vector])
processLine2 (r@(Rope (h : ks)), hist) (dir : _ : dist) =
  last . take (read dist + 1) . iterate (doStep dir) $ (r, hist)
processLine2 _ s = error $ "processline2 error: " ++ show s

doStep dir (r@(Rope (h : ks)), hist) =
  let newh = move h dir 1
      (newrope, newhist) = moveRope (Rope (newh : ks)) hist
   in (newrope, newhist)
doStep _ _ = error "wut"
