module Day10Lib
  ( --
    processInput1,
    processInput2,
  )
where

import Data.List
import Data.List.Split
import Debug.Trace

newtype CPU = CPU (Int, Int) deriving (Show, Eq)

data Instruction = AddX Int | Noop deriving (Show, Eq)

processInstruction :: Instruction -> CPU -> CPU
processInstruction (AddX n) (CPU (a, c)) = CPU (a + n, c + 2)
processInstruction Noop (CPU (a, c)) = CPU (a, c + 1)

parseInstruction :: String -> Instruction
parseInstruction ins = case words ins of
  ["addx", ns] -> AddX (read ns)
  ["noop"] -> Noop
  _ -> error $ "illegal instruction: " ++ ins

every :: Int -> [a] -> [a]
every n xs = case drop (n - 1) xs of
  y : ys -> y : every n ys
  [] -> []

-- ####################

processLine :: (CPU, [Int]) -> String -> (CPU, [Int])
processLine (cpu@(CPU (a, _)), hist) inss =
  let ins = parseInstruction inss
      newcpu = processInstruction ins cpu
   in (newcpu, hist ++ replicate (if ins == Noop then 1 else 2) a)

-- ####################

processInput1 :: String -> Int
processInput1 =
  sum
    . zipWith (*) [20, 60 ..]
    . every 40
    . (++) (replicate 20 0)
    . snd
    . foldl processLine (CPU (1, 0), [])
    . lines

processInput2 :: String -> String
processInput2 =
  intercalate "\n"
    . map (zipWith (\x y -> if abs ((y + 1) - x) <= 1 then '#' else '.') [1 ..])
    . chunksOf 40
    . snd
    . foldl processLine (CPU (1, 0), [])
    . lines
