module Day13Lib
  ( Value (..),
    parseList,
    parseNumber,
    --
    processInput1,
    processInput2,
  )
where

import Data.List
import Data.List.Split (splitOn)
import Debug.Trace
import Text.Parsec

data Value = Num Int | List [Value] deriving (Show, Eq)

instance Ord Value where
  compare (Num x) (Num y) = compare x y
  compare x@(Num _) y@(List _) = compare (List [x]) y
  compare x@(List _) y@(Num _) = compare x (List [y])
  compare (List (x : xs)) (List (y : ys)) = case compare x y of
    EQ -> compare (List xs) (List ys)
    o -> o
  compare (List []) (List (y : ys)) = LT
  compare (List (x : xs)) (List []) = GT
  compare (List []) (List []) = EQ

parseList :: Parsec String () Value
parseList = do
  char '['
  ls <- (parseNumber <|> parseList) `sepBy` char ','
  char ']'
  return $ List ls

parseNumber :: Parsec String () Value
parseNumber = do
  ns <- try . many1 $ digit
  return $ Num (read ns)

runParse :: String -> Value
runParse s = case parse parseList "" s of
  Left e -> error $ "parse error: " ++ show e
  Right v -> v

-- ####################

processInput1 :: String -> Int
processInput1 =
  sum
    . zipWith (\x [v1, v2] -> if v1 < v2 then x else 0) [1 ..]
    . map
      (map runParse . lines)
    . splitOn "\n\n"

processInput2 :: String -> Int
processInput2 =
  product
    . zipWith (\x y -> if y `elem` dividers then x else 1) [1 ..]
    . sort
    . (++) dividers
    . (concatMap (map runParse . lines) . splitOn "\n\n")

dividers = [runParse "[[2]]", runParse "[[6]]"]